<?php
// $Id: adserve.php,v 1.2.2.5 2007/01/02 05:47:09 jeremy Exp $

$debug = (int)$_GET['debug'];
$group = $_GET['g'];
$cache = $_GET['c'];
$path = $_GET['p'];
$files = $_GET['f'];
$hostid = $_GET['k'];

// Set defaults.
if (!$group) {
  $group = 'default';
}
if (!$cache) {
  $cache = 'database';
}

if ($debug) {
  echo "Group: '$group'<br />\n";
  echo "Cache: '$cache'<br />\n";
  echo "Host: '$hostid'<br />\n";
  echo "Path: '$path'<br />\n";
  echo "Files: '$files'<br />\n";
  if ($debug == 1) exit;
}

$ad_dir = getcwd();

// Move process to top level Drupal directory so we can perform a bootstrap.
for ($i = 1; $i <= 20; $i++) {
  if (file_exists('./includes/bootstrap.inc')) {
    break;
  }
  chdir('..');
}
$root_dir = getcwd();
require_once './includes/bootstrap.inc';

if ($debug) {
  echo "Ad module installed in '". $ad_dir ."'.<br />\n";
  echo "Root drupal directory '". $root_dir ."'.<br />\n";
}

switch ($cache) {
  case 'database':
    if ($debug) {
      echo "Using database cache.<br />\n";
    }

    adserve_bootstrap(DRUPAL_BOOTSTRAP_DATABASE);
    include_once './includes/module.inc';
  
    // Query to randomly determine which ad to display.
    if ($debug) {
      echo "SELECT a.aid, a.adtype, a.redirect FROM {ads} a INNER JOIN {ad_groups} g ON a.gid = g.gid WHERE a.adstatus = 'active' AND g.name = '$group' ORDER BY RAND() LIMIT 1<br />\n";
    }
    $ad = db_fetch_object(db_query("SELECT a.aid, a.adtype, a.redirect FROM {ads} a INNER JOIN {ad_groups} g ON a.gid = g.gid WHERE a.adstatus = 'active' AND g.name = '%s' ORDER BY RAND() LIMIT 1", $group));
    if (empty($ad)) {
      echo "No ad found.";
      return;
    }
    if ($debug) {
      echo "ad: <pre>";
      print_r($ad);
      echo "</pre>";
    }
    
    // Include appropriate module for displaying selected ad.
    $module = db_result(db_query("SELECT filename FROM {system} WHERE name = '%s'", "ad_$ad->adtype"));
    if ($debug) {
      echo "Loading module '$module'.<br />\n";
    }
    include_once "$module";
    
    $output = module_invoke("ad_$ad->adtype", 'display_ad', $ad, $hostid);

    if ($debug) {
      exit;
    }
  
    // Update view statistics.
    db_query("UPDATE {ad_statistics} SET count = count + 1 WHERE aid = %d AND action = 'view' AND date = %d AND hostid = '%s'", $ad->aid, date('YmdH'), $hostid);
    // If column doesn't already exist, we need to add it.
    if (!db_affected_rows()) {
      db_query("INSERT INTO {ad_statistics} (aid, date, action, hostid, count) VALUES(%d, %d, 'view', '%s', 1)", $ad->aid, date('YmdH'), $hostid);
      // If another process already added this row our INSERT will fail, if so 
      // we still need to increment it so we don't loose a view.
      if (!db_affected_rows()) {
        db_query("UPDATE {ad_statistics} SET count = count + 1 WHERE aid = %d AND action = 'view' AND date = %d AND hostid = '%s'", $ad->aid, date('YmdH'), $hostid);
      }
    }
  break;

  case 'file':
    if ($debug) {
      echo "Using file cache.<br />\n";
    }

    // We'll loop through all possible cache files until we obtain an exclusive
    // lock.
    $lock = FALSE;
    for ($i = 1; $i <= $files; $i++) {
      // Prefix the filename with a '.' to hide it on Unix systems.
      $cache_file = $path .'/.'. $i .'.ad.cache';

      if ($debug) {
        echo "Trying cache_file '$cache_file'.<br />\n";
      }

      if (!$fd = @fopen($cache_file, 'r+')) {
        if ($debug) {
          echo "Failed to open cache_file '$cache_file'.<br />\n";
        }
        // We failed to open the cache file, try the next one.
        continue;
      }
      if ($i < $files) {
        // This isn't the last available cache file so we'll use a non-blocking
        // lock for best performance.  If we fail to lock this cache file, we'll
        // quickly move on to the next until we find an available one.
        if (!flock($fd, LOCK_EX|LOCK_NB)) {
          if ($debug) {
            echo "Failed to obtain non-blocking lock.<br />\n";
          }
          // We failed to obtain an exclusive lock, close the file and try the
          // next one.
          @fclose($fd);
          continue;
        }
        if ($debug) {
          echo "Obtained lock.<br />\n";
        }
        $lock = TRUE;
        break;
      }
      else {
        // This is the last available cache file, we'll use a blocking lock as 
        // we have to wait until we have exclusive write permissions.
        if (!flock($fd, LOCK_EX)) {
          if ($debug) {
            echo "Failed to obtain blocking lock.<br />\n";
          }
          // A blocking exclusive lock shouldn't ever fail, so something has
          // gone very wrong.  Perhaps the file was deleted out from under us?
          @fclose($fd);
          continue;
        }
        if ($debug) {
          echo "Obtained lock.<br />\n";
        }
        $lock = TRUE;
        break;
      }
    }

    if ($lock) {
      // Read entire cache file into memory.
      $cache = unserialize(fread($fd, filesize($cache_file)));

      /**
       * The cache structure looks like this:
       * $cache['ad'][$group][]['display'] = $ad
       * $cache['ad'][$group][]['aid'] = $aid
       * $cache['ad'][$group][][$action][$timestamp] = $counter
       * $cache['last_sync'] = $timestamp
       */

      // Randomly select a banner to display.  We subtract 1 as arrays start
      // at 0.
      $cache_size = sizeof($cache['ad'][$group]);
      $id = $cache_size > 1 ? mt_rand(1, $cache_size) - 1 : 0;
      $aid = $cache['ad'][$group][$id]['aid'];
      $ad = $cache['ad'][$group][$id];
      $last_sync = $cache['last_sync'];

      if ($debug) {
        echo "Cache size: $cache_size<br />\n";
        echo "Random ID: $id<br />\n";
        echo "Ad ID: $aid<br />\n";
        echo "Last sync: $last_sync<br />\n";
      }

      $timestamp = date('YmdH');

      // Increment view counter.
      $cache['ad'][$group][$id]['view'][$timestamp]++;

      // Write updated cache back to file and release the lock.
      $cache = serialize($cache);
      rewind($fd);
      fwrite($fd, $cache, strlen($cache));
      flock($fd, LOCK_UN);
      fclose($fd);

      // Every 60 seconds we flush the cache files to the database.
      if ($last_sync < time() - 60) {
        adserve_rebuild_cache_files();
      }

      $output = $ad['display'];
    }
    else {
      $output = 'Failed to load image from cache file.';
      adserve_rebuild_cache_files();
    }
    break;

  default:
    /**
     * It's possible to define your own caching mechanism.  You simply have
     * to put all necessary logic in a file named "ad_cache_TYPE.module" where
     * TYPE is your type of cache.  For example, if you had a "memory" cache
     * you would create a file called "ad_cache_memory.module".  It must have a
     * function called ad_cache_TYPE(), so with our "memory" example above
     * the function would be named "ad_cache_memory()".  That function in
     * your cache file will then be called, and will have access to all the
     * global variables defined at the top of this file. (ie $group, $hostid,
     * etc...)  Your function must return the ad exactly as it is to be
     * displayed.
     */
    $type = "ad_cache_$cache";
    if (file_exists("$ad_dir/$type.module")) {
      require_once "$ad_dir/$type.module";
      if (function_exists($type)) {
        $output = $type();
      }
      else {
        $output = "Required function $type() does not exist.";
      }
    }
    else {
      $output = "Failed to include file $ad_dir/$type.module.";
    }
}

header("Content-Type: application/x-javascript; charset=utf-8");
print "document.write('". $output ."')";
exit(0);


/**
 * Bootstrap drupal, then run _ad_build_cache_files() from ad.module which will 
 * rebuild all cache files.
 */
function adserve_rebuild_cache_files() {
  adserve_bootstrap();
  _ad_build_cache_files();
}

/**
 * Include the necessary files and call the Drupal bootstrap.
 */
function adserve_bootstrap($bootstrap = DRUPAL_BOOTSTRAP_FULL) {
  global $debug;
  if ($debug) {
    echo "Drupal bootstrap '". $bootstrap ."'.<br />\n";
  }

  drupal_bootstrap($bootstrap);
}

?>
